@@ -16,9 +16,9 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library. If not, see <http://www.gnu.org/licenses/>.
+ *
+ *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -118,7 +118,7 @@ initialize_texture_units (CoglGles2Wrapper *w)
   /* We will need to set the matrix mode to GL_TEXTURE to
    * initialise any new texture units, so we save the current
    * mode for restoring afterwards */
-  GE( cogl_wrap_glGetIntegerv (GL_MATRIX_MODE, &prev_mode));
+  GE( _cogl_wrap_glGetIntegerv (GL_MATRIX_MODE, &prev_mode));
 
   for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
     {
@@ -128,11 +128,39 @@ initialize_texture_units (CoglGles2Wrapper *w)
       memset (new_unit, 0, sizeof (CoglGles2WrapperTextureUnit));
 
       w->active_texture_unit = i;
-      GE( cogl_wrap_glMatrixMode (GL_TEXTURE));
-      GE( cogl_wrap_glLoadIdentity ());
+      GE( _cogl_wrap_glMatrixMode (GL_TEXTURE));
+      GE( _cogl_wrap_glLoadIdentity ());
+
+      /* The real GL default is GL_MODULATE but the shader only
+         supports GL_COMBINE so let's default to that instead */
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,
+                                GL_COMBINE) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_RGB,
+                                GL_MODULATE) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_SRC0_RGB,
+                                GL_PREVIOUS) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_SRC1_RGB,
+                                GL_TEXTURE) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_RGB,
+                                GL_SRC_COLOR) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_RGB,
+                                GL_SRC_COLOR) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_COMBINE_ALPHA,
+                                GL_MODULATE) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_SRC0_ALPHA,
+                                GL_PREVIOUS) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_SRC1_ALPHA,
+                                GL_TEXTURE) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND0_ALPHA,
+                                GL_SRC_COLOR) );
+      GE( _cogl_wrap_glTexEnvi (GL_TEXTURE_ENV, GL_OPERAND1_ALPHA,
+                                GL_SRC_COLOR) );
+
+      GE( _cogl_wrap_glTexEnvi (GL_POINT_SPRITE, GL_COORD_REPLACE,
+                                GL_FALSE) );
     }
 
-  GE( cogl_wrap_glMatrixMode ((GLenum) prev_mode));
+  GE( _cogl_wrap_glMatrixMode ((GLenum) prev_mode));
 
   w->settings.texture_units = 0;
 
@@ -140,44 +168,70 @@ initialize_texture_units (CoglGles2Wrapper *w)
 }
 
 void
-cogl_gles2_wrapper_init (CoglGles2Wrapper *wrapper)
+_cogl_gles2_wrapper_init (CoglGles2Wrapper *wrapper)
 {
   GLfloat default_fog_color[4] = { 0, 0, 0, 0 };
 
   memset (wrapper, 0, sizeof (CoglGles2Wrapper));
 
   /* Initialize the stacks */
-  cogl_wrap_glMatrixMode (GL_PROJECTION);
-  cogl_wrap_glLoadIdentity ();
-  cogl_wrap_glMatrixMode (GL_MODELVIEW);
-  cogl_wrap_glLoadIdentity ();
+  _cogl_wrap_glMatrixMode (GL_PROJECTION);
+  _cogl_wrap_glLoadIdentity ();
+  _cogl_wrap_glMatrixMode (GL_MODELVIEW);
+  _cogl_wrap_glLoadIdentity ();
 
   /* The gl*ActiveTexture wrappers will initialise the texture
    * stack for the texture unit when it's first activated */
-  cogl_wrap_glActiveTexture (GL_TEXTURE0);
-  cogl_wrap_glClientActiveTexture (GL_TEXTURE0);
+  _cogl_wrap_glActiveTexture (GL_TEXTURE0);
+  _cogl_wrap_glClientActiveTexture (GL_TEXTURE0);
 
   /* Initialize the fogging options */
-  cogl_wrap_glDisable (GL_FOG);
-  cogl_wrap_glFogf (GL_FOG_MODE, GL_LINEAR);
-  cogl_wrap_glFogf (GL_FOG_DENSITY, 1.0);
-  cogl_wrap_glFogf (GL_FOG_START, 0);
-  cogl_wrap_glFogf (GL_FOG_END, 1);
-  cogl_wrap_glFogfv (GL_FOG_COLOR, default_fog_color);
+  _cogl_wrap_glDisable (GL_FOG);
+  _cogl_wrap_glFogf (GL_FOG_MODE, GL_LINEAR);
+  _cogl_wrap_glFogf (GL_FOG_DENSITY, 1.0);
+  _cogl_wrap_glFogf (GL_FOG_START, 0);
+  _cogl_wrap_glFogf (GL_FOG_END, 1);
+  _cogl_wrap_glFogfv (GL_FOG_COLOR, default_fog_color);
 
   /* Initialize alpha testing */
-  cogl_wrap_glDisable (GL_ALPHA_TEST);
-  cogl_wrap_glAlphaFunc (GL_ALWAYS, 0.0f);
+  _cogl_wrap_glDisable (GL_ALPHA_TEST);
+  _cogl_wrap_glAlphaFunc (GL_ALWAYS, 0.0f);
+
+  /* Initialize the point size */
+  _cogl_wrap_glPointSize (1.0f);
 
   initialize_texture_units (wrapper);
 }
 
+static int
+cogl_gles2_get_n_args_for_combine_func (GLenum func)
+{
+  switch (func)
+    {
+    case GL_REPLACE:
+      return 1;
+    case GL_MODULATE:
+    case GL_ADD:
+    case GL_ADD_SIGNED:
+    case GL_SUBTRACT:
+    case GL_DOT3_RGB:
+    case GL_DOT3_RGBA:
+      return 2;
+    case GL_INTERPOLATE:
+      return 3;
+    default:
+      return 0;
+    }
+}
+
 static gboolean
 cogl_gles2_settings_equal (const CoglGles2WrapperSettings *a,
 			   const CoglGles2WrapperSettings *b,
 			   gboolean vertex_tests,
 			   gboolean fragment_tests)
 {
+  int i;
+
   if (a->texture_units != b->texture_units)
     return FALSE;
 
@@ -195,6 +249,38 @@ cogl_gles2_settings_equal (const CoglGles2WrapperSettings *a,
   if (vertex_tests && a->fog_enabled && a->fog_mode != b->fog_mode)
     return FALSE;
 
+  /* Compare layer combine operation for each active unit */
+  if (fragment_tests)
+    for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
+      if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (a->texture_units, i))
+        {
+          const CoglGles2WrapperTexEnv *tex_env_a = a->tex_env + i;
+          const CoglGles2WrapperTexEnv *tex_env_b = b->tex_env + i;
+          int arg, n_args;
+          GLenum func;
+
+          if (tex_env_a->point_sprite_coords !=
+              tex_env_b->point_sprite_coords)
+            return FALSE;
+
+          func = tex_env_a->texture_combine_rgb_func;
+
+          if (func != tex_env_b->texture_combine_rgb_func)
+            return FALSE;
+
+          n_args = cogl_gles2_get_n_args_for_combine_func (func);
+
+          for (arg = 0; arg < n_args; arg++)
+            {
+              if (tex_env_a->texture_combine_rgb_src[arg] !=
+                  tex_env_b->texture_combine_rgb_src[arg])
+                return FALSE;
+              if (tex_env_a->texture_combine_rgb_op[arg] !=
+                  tex_env_b->texture_combine_rgb_op[arg])
+                return FALSE;
+            }
+        }
+
   return TRUE;
 }
 
@@ -219,7 +305,7 @@ cogl_gles2_get_vertex_shader (const CoglGles2WrapperSettings *settings)
       return (CoglGles2WrapperShader *) node->data;
 
   /* Otherwise create a new shader */
-  shader_source = g_string_new (cogl_fixed_vertex_shader_per_vertex_attribs);
+  shader_source = g_string_new (_cogl_fixed_vertex_shader_per_vertex_attribs);
 
   for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
     if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
@@ -232,8 +318,8 @@ cogl_gles2_get_vertex_shader (const CoglGles2WrapperSettings *settings)
     if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
       n_texture_units = i + 1;
 
-  g_string_append (shader_source, cogl_fixed_vertex_shader_transform_matrices);
-  g_string_append (shader_source, cogl_fixed_vertex_shader_output_variables);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_transform_matrices);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_output_variables);
 
   if (n_texture_units > 0)
     {
@@ -246,11 +332,12 @@ cogl_gles2_get_vertex_shader (const CoglGles2WrapperSettings *settings)
                               n_texture_units);
     }
 
-  g_string_append (shader_source, cogl_fixed_vertex_shader_fogging_options);
-  g_string_append (shader_source, cogl_fixed_vertex_shader_main_start);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_fogging_options);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_main_start);
 
   for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
-    if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
+    if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i) &&
+        !settings->tex_env[i].point_sprite_coords)
       {
         g_string_append_printf (shader_source,
                                 "transformed_tex_coord = "
@@ -263,31 +350,31 @@ cogl_gles2_get_vertex_shader (const CoglGles2WrapperSettings *settings)
                                 i);
       }
 
-  g_string_append (shader_source, cogl_fixed_vertex_shader_frag_color_start);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_frag_color_start);
 
   if (settings->fog_enabled)
     {
-      g_string_append (shader_source, cogl_fixed_vertex_shader_fog_start);
+      g_string_append (shader_source, _cogl_fixed_vertex_shader_fog_start);
 
       switch (settings->fog_mode)
 	{
 	case GL_EXP:
-	  g_string_append (shader_source, cogl_fixed_vertex_shader_fog_exp);
+          g_string_append (shader_source, _cogl_fixed_vertex_shader_fog_exp);
 	  break;
 
 	case GL_EXP2:
-	  g_string_append (shader_source, cogl_fixed_vertex_shader_fog_exp2);
+          g_string_append (shader_source, _cogl_fixed_vertex_shader_fog_exp2);
 	  break;
 
 	default:
-	  g_string_append (shader_source, cogl_fixed_vertex_shader_fog_linear);
+          g_string_append (shader_source, _cogl_fixed_vertex_shader_fog_linear);
 	  break;
 	}
 
-      g_string_append (shader_source, cogl_fixed_vertex_shader_fog_end);
+      g_string_append (shader_source, _cogl_fixed_vertex_shader_fog_end);
     }
 
-  g_string_append (shader_source, cogl_fixed_vertex_shader_end);
+  g_string_append (shader_source, _cogl_fixed_vertex_shader_end);
 
   shader_obj = cogl_gles2_wrapper_create_shader (GL_VERTEX_SHADER,
 						 shader_source->str);
@@ -307,6 +394,201 @@ cogl_gles2_get_vertex_shader (const CoglGles2WrapperSettings *settings)
   return shader;
 }
 
+static void
+cogl_gles2_add_texture_lookup (int unit,
+                               const char *swizzle,
+                               GString *shader_source)
+{
+  _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
+
+  g_string_append_printf (shader_source, "texture2D (texture_unit[%d], ", unit);
+
+  /* If point sprite coord generation is being used then divert to the
+     built-in varying var for that instead of the texture
+     coordinates */
+  if (w->settings.tex_env[unit].point_sprite_coords)
+    g_string_append (shader_source, "gl_PointCoord");
+  else
+    g_string_append_printf (shader_source, "tex_coord[%d]", unit);
+
+  g_string_append_printf (shader_source, ").%s", swizzle);
+}
+
+static void
+cogl_gles2_add_arg (int unit,
+                    GLenum src,
+                    GLenum operand,
+                    const char *swizzle,
+                    GString *shader_source)
+{
+  char alpha_swizzle[5] = "aaaa";
+
+  g_string_append_c (shader_source, '(');
+
+  if (operand == GL_ONE_MINUS_SRC_COLOR || operand == GL_ONE_MINUS_SRC_ALPHA)
+    g_string_append_printf (shader_source,
+                            "vec4(1.0, 1.0, 1.0, 1.0).%s - ",
+                            swizzle);
+
+  /* If the operand is reading from the alpha then replace the swizzle
+     with the same number of copies of the alpha */
+  if (operand == GL_SRC_ALPHA || operand == GL_ONE_MINUS_SRC_ALPHA)
+    {
+      alpha_swizzle[strlen (swizzle)] = '\0';
+      swizzle = alpha_swizzle;
+    }
+
+  switch (src)
+    {
+    case GL_TEXTURE:
+      cogl_gles2_add_texture_lookup (unit, swizzle, shader_source);
+      break;
+
+    case GL_CONSTANT:
+      g_string_append_printf (shader_source, "combine_constant[%d].%s",
+                              unit, swizzle);
+      break;
+
+    case GL_PREVIOUS:
+      if (unit > 0)
+        {
+          g_string_append_printf (shader_source, "gl_FragColor.%s", swizzle);
+          break;
+        }
+      /* flow through */
+    case GL_PRIMARY_COLOR:
+      g_string_append_printf (shader_source, "frag_color.%s", swizzle);
+      break;
+
+    default:
+      if (src >= GL_TEXTURE0 &&
+          src < GL_TEXTURE0 + COGL_GLES2_MAX_TEXTURE_UNITS)
+        cogl_gles2_add_texture_lookup (src - GL_TEXTURE0,
+                                       swizzle, shader_source);
+      break;
+    }
+
+  g_string_append_c (shader_source, ')');
+}
+
+static void
+cogl_gles2_add_operation (int unit,
+                          GLenum combine_func,
+                          const GLenum *sources,
+                          const GLenum *operands,
+                          const char *swizzle,
+                          GString *shader_source)
+{
+  switch (combine_func)
+    {
+    case GL_REPLACE:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      break;
+
+    case GL_MODULATE:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " * ");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          swizzle, shader_source);
+      break;
+
+    case GL_ADD:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " + ");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          swizzle, shader_source);
+      break;
+
+    case GL_ADD_SIGNED:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " + ");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          swizzle, shader_source);
+      g_string_append_printf (shader_source,
+                              " - vec4(0.5, 0.5, 0.5, 0.5).%s",
+                              swizzle);
+      break;
+
+    case GL_SUBTRACT:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " - ");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          swizzle, shader_source);
+      break;
+
+    case GL_INTERPOLATE:
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " * ");
+      cogl_gles2_add_arg (unit, sources[2], operands[2],
+                          swizzle, shader_source);
+      g_string_append (shader_source, " + ");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          swizzle, shader_source);
+      g_string_append_printf (shader_source,
+                              " * (vec4(1.0, 1.0, 1.0, 1.0).%s - ",
+                              swizzle);
+      cogl_gles2_add_arg (unit, sources[2], operands[2],
+                          swizzle, shader_source);
+      g_string_append_c (shader_source, ')');
+      break;
+
+    case GL_DOT3_RGB:
+    case GL_DOT3_RGBA:
+      g_string_append (shader_source, "vec4(4 * ((");
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          "r", shader_source);
+      g_string_append (shader_source, " - 0.5) * (");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          "r", shader_source);
+      g_string_append (shader_source, " - 0.5) + (");
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          "g", shader_source);
+      g_string_append (shader_source, " - 0.5) * (");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          "g", shader_source);
+      g_string_append (shader_source, " - 0.5) + (");
+      cogl_gles2_add_arg (unit, sources[0], operands[0],
+                          "b", shader_source);
+      g_string_append (shader_source, " - 0.5) * (");
+      cogl_gles2_add_arg (unit, sources[1], operands[1],
+                          "b", shader_source);
+      g_string_append_printf (shader_source, " - 0.5))).%s", swizzle);
+      break;
+    }
+}
+
+static gboolean
+cogl_gles2_rgb_and_alpha_equal (const CoglGles2WrapperTexEnv *tex_env)
+{
+  int arg, n_args;
+
+  if (tex_env->texture_combine_rgb_func != tex_env->texture_combine_alpha_func)
+    return FALSE;
+
+  n_args =
+    cogl_gles2_get_n_args_for_combine_func (tex_env->texture_combine_rgb_func);
+
+  for (arg = 0; arg < n_args; arg++)
+    {
+      if (tex_env->texture_combine_rgb_src[arg] !=
+          tex_env->texture_combine_alpha_src[arg])
+        return FALSE;
+
+      if (tex_env->texture_combine_rgb_op[arg] != GL_SRC_COLOR ||
+          tex_env->texture_combine_alpha_op[arg] != GL_SRC_ALPHA)
+        return FALSE;
+    }
+
+  return TRUE;
+}
+
+
 static CoglGles2WrapperShader *
 cogl_gles2_get_fragment_shader (const CoglGles2WrapperSettings *settings)
 {
@@ -328,14 +610,14 @@ cogl_gles2_get_fragment_shader (const CoglGles2WrapperSettings *settings)
       return (CoglGles2WrapperShader *) node->data;
 
   /* Otherwise create a new shader */
-  shader_source = g_string_new (cogl_fixed_fragment_shader_variables_start);
+  shader_source = g_string_new (_cogl_fixed_fragment_shader_variables_start);
 
   /* Find the biggest enabled texture unit index */
   for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
     if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
       n_texture_units = i + 1;
 
-  g_string_append (shader_source, cogl_fixed_fragment_shader_inputs);
+  g_string_append (shader_source, _cogl_fixed_fragment_shader_inputs);
 
   if (n_texture_units > 0)
     {
@@ -343,82 +625,109 @@ cogl_gles2_get_fragment_shader (const CoglGles2WrapperSettings *settings)
                               "varying vec2       tex_coord[%d];\n",
                               n_texture_units);
 
-      g_string_append (shader_source, cogl_fixed_fragment_shader_texturing_options);
+      g_string_append (shader_source,
+                       _cogl_fixed_fragment_shader_texturing_options);
       g_string_append_printf (shader_source,
                               "uniform sampler2D  texture_unit[%d];\n",
                               n_texture_units);
     }
 
-  g_string_append (shader_source, cogl_fixed_fragment_shader_fogging_options);
+  g_string_append (shader_source, _cogl_fixed_fragment_shader_fogging_options);
 
-  g_string_append (shader_source, cogl_fixed_fragment_shader_main_declare);
+  g_string_append (shader_source, _cogl_fixed_fragment_shader_main_declare);
 
-  g_string_append (shader_source, cogl_fixed_fragment_shader_main_start);
+  g_string_append (shader_source, _cogl_fixed_fragment_shader_main_start);
 
   /* This pointless extra variable is needed to work around an
      apparent bug in the PowerVR drivers. Without it the alpha
      blending seems to stop working */
   g_string_append (shader_source,
-		   "vec4 frag_color_copy = frag_color;\n");
-  g_string_append (shader_source, "gl_FragColor = frag_color;\n");
+                   "vec4 frag_color_copy = frag_color;\n");
 
-  for (i = 0; i < n_texture_units; i++)
-    if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
-      {
-        if (COGL_GLES2_TEXTURE_UNIT_IS_ALPHA_ONLY (settings->texture_units, i))
-          /* If the texture only has an alpha channel (eg, with the textures
-             from the pango renderer) then the RGB components will be
-             black. We want to use the RGB from the current color in that
-             case */
-          g_string_append_printf (shader_source,
-                                  "gl_FragColor.a *= "
-                                  "texture2D (texture_unit[%d], "
-                                  "tex_coord[%d]).a;\n",
-                                  i, i);
-        else
-          g_string_append_printf (shader_source,
-                                  "gl_FragColor *= "
-                                  "texture2D (texture_unit[%d], "
-                                  "tex_coord[%d]);\n",
-                                  i, i);
-      }
+  /* If there are no textures units enabled then we can just directly
+     use the color from the vertex shader */
+  if (n_texture_units == 0)
+    g_string_append (shader_source, "gl_FragColor = frag_color;\n");
+  else
+    /* Otherwise we need to calculate the value based on the layer
+       combine settings */
+    for (i = 0; i < n_texture_units; i++)
+      if (COGL_GLES2_TEXTURE_UNIT_IS_ENABLED (settings->texture_units, i))
+        {
+          const CoglGles2WrapperTexEnv *tex_env = settings->tex_env + i;
+
+          /* If the rgb and alpha combine functions are the same then
+             we can do both with a single statement, otherwise we need
+             to do them separately */
+          if (cogl_gles2_rgb_and_alpha_equal (tex_env))
+            {
+              g_string_append (shader_source, "gl_FragColor.rgba = ");
+              cogl_gles2_add_operation (i,
+                                        tex_env->texture_combine_rgb_func,
+                                        tex_env->texture_combine_rgb_src,
+                                        tex_env->texture_combine_rgb_op,
+                                        "rgba",
+                                        shader_source);
+              g_string_append (shader_source, ";\n");
+            }
+          else
+            {
+              g_string_append (shader_source, "gl_FragColor.rgb = ");
+              cogl_gles2_add_operation (i,
+                                        tex_env->texture_combine_rgb_func,
+                                        tex_env->texture_combine_rgb_src,
+                                        tex_env->texture_combine_rgb_op,
+                                        "rgb",
+                                        shader_source);
+              g_string_append (shader_source,
+                               ";\n"
+                               "gl_FragColor.a = ");
+              cogl_gles2_add_operation (i,
+                                        tex_env->texture_combine_alpha_func,
+                                        tex_env->texture_combine_alpha_src,
+                                        tex_env->texture_combine_alpha_op,
+                                        "a",
+                                        shader_source);
+              g_string_append (shader_source, ";\n");
+            }
+        }
 
   if (settings->fog_enabled)
-    g_string_append (shader_source, cogl_fixed_fragment_shader_fog);
+    g_string_append (shader_source, _cogl_fixed_fragment_shader_fog);
 
   if (settings->alpha_test_enabled)
     switch (settings->alpha_test_func)
       {
       case GL_NEVER:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_never);
+                         _cogl_fixed_fragment_shader_alpha_never);
 	break;
       case GL_LESS:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_less);
+                         _cogl_fixed_fragment_shader_alpha_less);
 	break;
       case GL_EQUAL:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_equal);
+                         _cogl_fixed_fragment_shader_alpha_equal);
 	break;
       case GL_LEQUAL:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_lequal);
+                         _cogl_fixed_fragment_shader_alpha_lequal);
 	break;
       case GL_GREATER:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_greater);
+                         _cogl_fixed_fragment_shader_alpha_greater);
 	break;
       case GL_NOTEQUAL:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_notequal);
+                         _cogl_fixed_fragment_shader_alpha_notequal);
 	break;
       case GL_GEQUAL:
 	g_string_append (shader_source,
-			 cogl_fixed_fragment_shader_alpha_gequal);
+                         _cogl_fixed_fragment_shader_alpha_gequal);
       }
 
-  g_string_append (shader_source, cogl_fixed_fragment_shader_end);
+  g_string_append (shader_source, _cogl_fixed_fragment_shader_end);
 
   shader_obj = cogl_gles2_wrapper_create_shader (GL_FRAGMENT_SHADER,
 						 shader_source->str);
@@ -488,6 +797,9 @@ cogl_gles2_wrapper_get_locations (GLuint program,
 
   uniforms->alpha_test_ref_uniform
     = glGetUniformLocation (program, "alpha_test_ref");
+
+  uniforms->point_size_uniform
+    = glGetUniformLocation (program, "point_size");
 }
 
 static void
@@ -613,7 +925,7 @@ cogl_gles2_wrapper_get_program (const CoglGles2WrapperSettings *settings)
 }
 
 void
-cogl_gles2_wrapper_deinit (CoglGles2Wrapper *wrapper)
+_cogl_gles2_wrapper_deinit (CoglGles2Wrapper *wrapper)
 {
   GSList *node, *next;
   int i;
@@ -676,7 +988,7 @@ cogl_gles2_wrapper_notify_matrix_changed (CoglGles2Wrapper *wrapper,
 }
 
 void
-cogl_wrap_glMatrixMode (GLenum mode)
+_cogl_wrap_glMatrixMode (GLenum mode)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -708,7 +1020,7 @@ cogl_gles2_get_current_matrix (CoglGles2Wrapper *wrapper)
 }
 
 void
-cogl_wrap_glLoadIdentity (void)
+_cogl_wrap_glLoadIdentity (void)
 {
   CoglMatrix *matrix;
 
@@ -722,7 +1034,7 @@ cogl_wrap_glLoadIdentity (void)
 }
 
 void
-cogl_wrap_glLoadMatrixf (const GLfloat *m)
+_cogl_wrap_glLoadMatrixf (const GLfloat *m)
 {
   CoglMatrix *matrix;
 
@@ -736,7 +1048,7 @@ cogl_wrap_glLoadMatrixf (const GLfloat *m)
 }
 
 void
-cogl_wrap_glVertexPointer (GLint size, GLenum type, GLsizei stride,
+_cogl_wrap_glVertexPointer (GLint size, GLenum type, GLsizei stride,
 			   const GLvoid *pointer)
 {
   glVertexAttribPointer (COGL_GLES2_WRAPPER_VERTEX_ATTRIB, size, type,
@@ -744,7 +1056,7 @@ cogl_wrap_glVertexPointer (GLint size, GLenum type, GLsizei stride,
 }
 
 void
-cogl_wrap_glTexCoordPointer (GLint size, GLenum type, GLsizei stride,
+_cogl_wrap_glTexCoordPointer (GLint size, GLenum type, GLsizei stride,
 			     const GLvoid *pointer)
 {
   int active_unit;
@@ -764,15 +1076,15 @@ cogl_wrap_glTexCoordPointer (GLint size, GLenum type, GLsizei stride,
 }
 
 void
-cogl_wrap_glColorPointer (GLint size, GLenum type, GLsizei stride,
-			  const GLvoid *pointer)
+_cogl_wrap_glColorPointer (GLint size, GLenum type, GLsizei stride,
+                           const GLvoid *pointer)
 {
   glVertexAttribPointer (COGL_GLES2_WRAPPER_COLOR_ATTRIB, size, type,
 			 GL_TRUE, stride, pointer);
 }
 
 void
-cogl_wrap_glNormalPointer (GLenum type, GLsizei stride, const GLvoid *pointer)
+_cogl_wrap_glNormalPointer (GLenum type, GLsizei stride, const GLvoid *pointer)
 {
   glVertexAttribPointer (COGL_GLES2_WRAPPER_NORMAL_ATTRIB, 1, type,
 			 GL_FALSE, stride, pointer);
@@ -788,7 +1100,7 @@ cogl_gles2_do_set_uniform (GLint location, CoglBoxedValue *value)
 
     case COGL_BOXED_INT:
       {
-	gint *ptr;
+	int *ptr;
 
 	if (value->count == 1)
 	  ptr = value->v.int_value;
@@ -807,7 +1119,7 @@ cogl_gles2_do_set_uniform (GLint location, CoglBoxedValue *value)
 
     case COGL_BOXED_FLOAT:
       {
-	gfloat *ptr;
+	float *ptr;
 
 	if (value->count == 1)
 	  ptr = value->v.float_value;
@@ -826,7 +1138,7 @@ cogl_gles2_do_set_uniform (GLint location, CoglBoxedValue *value)
 
     case COGL_BOXED_MATRIX:
       {
-	gfloat *ptr;
+	float *ptr;
 
 	if (value->count == 1)
 	  ptr = value->v.matrix;
@@ -851,7 +1163,7 @@ cogl_gles2_do_set_uniform (GLint location, CoglBoxedValue *value)
 }
 
 static void
-cogl_wrap_prepare_for_draw (void)
+_cogl_wrap_prepare_for_draw (void)
 {
   CoglGles2WrapperProgram *program;
 
@@ -951,6 +1263,10 @@ cogl_wrap_prepare_for_draw (void)
             }
         }
 
+      if ((w->dirty_uniforms & COGL_GLES2_DIRTY_POINT_SIZE))
+        glUniform1f (program->uniforms.point_size_uniform,
+                     w->point_size);
+
       w->dirty_uniforms = 0;
     }
 
@@ -993,7 +1309,7 @@ cogl_wrap_prepare_for_draw (void)
             GLint tex_coord_var_index;
             CoglGles2WrapperTextureUnit *texture_unit;
 
-            texture_unit = w->texture_units + w->active_texture_unit;
+            texture_unit = w->texture_units + i;
             if (!texture_unit->texture_coords_enabled)
               continue;
 
@@ -1020,8 +1336,7 @@ cogl_wrap_prepare_for_draw (void)
 
       for (i = 0; i < COGL_GLES2_MAX_TEXTURE_UNITS; i++)
 	{
-	  CoglGles2WrapperTextureUnit *texture_unit
-            = w->texture_units + w->active_texture_unit;
+          CoglGles2WrapperTextureUnit *texture_unit = w->texture_units + i;
           GLint attrib = program->attributes.multi_texture_coords[i];
 
           if (attrib != -1)
@@ -1038,61 +1353,95 @@ cogl_wrap_prepare_for_draw (void)
 }
 
 void
-cogl_wrap_glDrawArrays (GLenum mode, GLint first, GLsizei count)
+_cogl_wrap_glDrawArrays (GLenum mode, GLint first, GLsizei count)
 {
-  cogl_wrap_prepare_for_draw ();
+  _cogl_wrap_prepare_for_draw ();
 
   glDrawArrays (mode, first, count);
 }
 
 void
-cogl_wrap_glDrawElements (GLenum mode, GLsizei count, GLenum type,
+_cogl_wrap_glDrawElements (GLenum mode, GLsizei count, GLenum type,
                           const GLvoid *indices)
 {
-  cogl_wrap_prepare_for_draw ();
+  _cogl_wrap_prepare_for_draw ();
 
   glDrawElements (mode, count, type, indices);
 }
 
 void
-cogl_gles2_wrapper_bind_texture (GLenum target, GLuint texture,
-				 GLenum internal_format)
+_cogl_wrap_glTexEnvi (GLenum target, GLenum pname, GLint param)
 {
+  CoglGles2WrapperTexEnv *tex_env;
+
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
-  glBindTexture (target, texture);
+  tex_env = w->settings.tex_env + w->active_texture_unit;
 
-  /* We need to keep track of whether the texture is alpha-only
-     because the emulation of GL_MODULATE needs to work differently in
-     that case */
-  if (COGL_GLES2_TEXTURE_UNIT_IS_ALPHA_ONLY (w->settings.texture_units,
-                                             w->active_texture_unit)
-      != (internal_format == GL_ALPHA))
+  if (target == GL_TEXTURE_ENV)
     {
-      COGL_GLES2_TEXTURE_UNIT_SET_ALPHA_ONLY (w->settings.texture_units,
-                                              w->active_texture_unit,
-                                              internal_format == GL_ALPHA);
+      switch (pname)
+        {
+        case GL_COMBINE_RGB:
+          tex_env->texture_combine_rgb_func = param;
+          break;
+        case GL_COMBINE_ALPHA:
+          tex_env->texture_combine_alpha_func = param;
+          break;
+        case GL_SRC0_RGB:
+        case GL_SRC1_RGB:
+        case GL_SRC2_RGB:
+          tex_env->texture_combine_rgb_src[pname - GL_SRC0_RGB] = param;
+          break;
+        case GL_SRC0_ALPHA:
+        case GL_SRC1_ALPHA:
+        case GL_SRC2_ALPHA:
+          tex_env->texture_combine_alpha_src[pname - GL_SRC0_ALPHA] = param;
+          break;
+        case GL_OPERAND0_RGB:
+        case GL_OPERAND1_RGB:
+        case GL_OPERAND2_RGB:
+          tex_env->texture_combine_rgb_op[pname - GL_OPERAND0_RGB] = param;
+          break;
+        case GL_OPERAND0_ALPHA:
+        case GL_OPERAND1_ALPHA:
+        case GL_OPERAND2_ALPHA:
+          tex_env->texture_combine_alpha_op[pname - GL_OPERAND0_ALPHA] = param;
+          break;
+        }
+
+      w->settings_dirty = TRUE;
+    }
+  else if (target == GL_POINT_SPRITE)
+    {
+      switch (pname)
+        {
+        case GL_COORD_REPLACE:
+          tex_env->point_sprite_coords = param;
+          break;
+        }
+
       w->settings_dirty = TRUE;
     }
 }
 
 void
-cogl_wrap_glTexEnvi (GLenum target, GLenum pname, GLint param)
+_cogl_wrap_glTexEnvfv (GLenum target, GLenum pname, const GLfloat *params)
 {
-  /* This function is only used to set the texture mode once to
-     GL_MODULATE. The shader is hard-coded to modulate the texture so
-     nothing needs to be done here. */
+  if (target == GL_TEXTURE_ENV && pname == GL_TEXTURE_ENV_COLOR)
+    {
+      CoglGles2WrapperTexEnv *tex_env;
+
+      _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
+
+      tex_env = w->settings.tex_env + w->active_texture_unit;
+
+      memcpy (tex_env->texture_combine_constant, params, sizeof (GLfloat) * 4);
+    }
 }
 
 void
-cogl_wrap_glTexEnvfv (GLenum target, GLenum pname, const GLfloat *params)
-{
-  /* FIXME: Currently needed to support texture combining using
-   * COGL_BLEND_STRING_COLOR_SOURCE_CONSTANT */
-}
-
-void
-cogl_wrap_glClientActiveTexture (GLenum texture)
+_cogl_wrap_glClientActiveTexture (GLenum texture)
 {
   int texture_unit_index = texture - GL_TEXTURE0;
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
@@ -1102,7 +1451,7 @@ cogl_wrap_glClientActiveTexture (GLenum texture)
 }
 
 void
-cogl_wrap_glActiveTexture (GLenum texture)
+_cogl_wrap_glActiveTexture (GLenum texture)
 {
   int texture_unit_index = texture - GL_TEXTURE0;
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
@@ -1114,7 +1463,7 @@ cogl_wrap_glActiveTexture (GLenum texture)
 }
 
 void
-cogl_wrap_glEnable (GLenum cap)
+_cogl_wrap_glEnable (GLenum cap)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -1145,7 +1494,7 @@ cogl_wrap_glEnable (GLenum cap)
 }
 
 void
-cogl_wrap_glDisable (GLenum cap)
+_cogl_wrap_glDisable (GLenum cap)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -1176,7 +1525,7 @@ cogl_wrap_glDisable (GLenum cap)
 }
 
 void
-cogl_wrap_glEnableClientState (GLenum array)
+_cogl_wrap_glEnableClientState (GLenum array)
 {
   CoglGles2WrapperTextureUnit *texture_unit;
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
@@ -1189,7 +1538,7 @@ cogl_wrap_glEnableClientState (GLenum array)
     case GL_TEXTURE_COORD_ARRAY:
       /* TODO - review if this should be in w->settings? */
 
-      texture_unit = w->texture_units + w->active_texture_unit;
+      texture_unit = w->texture_units + w->active_client_texture_unit;
       if (texture_unit->texture_coords_enabled != 1)
 	{
 	  texture_unit->texture_coords_enabled = 1;
@@ -1207,7 +1556,7 @@ cogl_wrap_glEnableClientState (GLenum array)
 }
 
 void
-cogl_wrap_glDisableClientState (GLenum array)
+_cogl_wrap_glDisableClientState (GLenum array)
 {
   CoglGles2WrapperTextureUnit *texture_unit;
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
@@ -1238,7 +1587,7 @@ cogl_wrap_glDisableClientState (GLenum array)
 }
 
 void
-cogl_wrap_glAlphaFunc (GLenum func, GLclampf ref)
+_cogl_wrap_glAlphaFunc (GLenum func, GLclampf ref)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -1252,26 +1601,26 @@ cogl_wrap_glAlphaFunc (GLenum func, GLclampf ref)
 }
 
 void
-cogl_wrap_glColor4f (GLclampf r, GLclampf g, GLclampf b, GLclampf a)
+_cogl_wrap_glColor4f (GLclampf r, GLclampf g, GLclampf b, GLclampf a)
 {
   glVertexAttrib4f (COGL_GLES2_WRAPPER_COLOR_ATTRIB, r, g, b, a);
 }
 
 void
-cogl_wrap_glColor4ub (GLubyte r, GLubyte g, GLubyte b, GLubyte a)
+_cogl_wrap_glColor4ub (GLubyte r, GLubyte g, GLubyte b, GLubyte a)
 {
   glVertexAttrib4f (COGL_GLES2_WRAPPER_COLOR_ATTRIB,
                     r/255.0, g/255.0, b/255.0, a/255.0);
 }
 
 void
-cogl_wrap_glClipPlanef (GLenum plane, GLfloat *equation)
+_cogl_wrap_glClipPlanef (GLenum plane, GLfloat *equation)
 {
   /* FIXME */
 }
 
 void
-cogl_wrap_glGetIntegerv (GLenum pname, GLint *params)
+_cogl_wrap_glGetIntegerv (GLenum pname, GLint *params)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -1285,6 +1634,12 @@ cogl_wrap_glGetIntegerv (GLenum pname, GLint *params)
       *params = w->matrix_mode;
       break;
 
+    case GL_MAX_TEXTURE_UNITS:
+      glGetIntegerv (GL_MAX_TEXTURE_IMAGE_UNITS, params);
+      if (*params > COGL_GLES2_MAX_TEXTURE_UNITS)
+        *params = COGL_GLES2_MAX_TEXTURE_UNITS;
+      break;
+
     default:
       glGetIntegerv (pname, params);
       break;
@@ -1292,7 +1647,7 @@ cogl_wrap_glGetIntegerv (GLenum pname, GLint *params)
 }
 
 void
-cogl_wrap_glGetFloatv (GLenum pname, GLfloat *params)
+_cogl_wrap_glGetFloatv (GLenum pname, GLfloat *params)
 {
   CoglGles2WrapperTextureUnit *texture_unit;
 
@@ -1320,7 +1675,7 @@ cogl_wrap_glGetFloatv (GLenum pname, GLfloat *params)
 }
 
 void
-cogl_wrap_glFogf (GLenum pname, GLfloat param)
+_cogl_wrap_glFogf (GLenum pname, GLfloat param)
 {
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
 
@@ -1348,7 +1703,7 @@ cogl_wrap_glFogf (GLenum pname, GLfloat param)
 }
 
 void
-cogl_wrap_glFogfv (GLenum pname, const GLfloat *params)
+_cogl_wrap_glFogfv (GLenum pname, const GLfloat *params)
 {
   int i;
   _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
@@ -1363,19 +1718,28 @@ cogl_wrap_glFogfv (GLenum pname, const GLfloat *params)
 }
 
 void
-cogl_wrap_glTexParameteri (GLenum target, GLenum pname, GLfloat param)
+_cogl_wrap_glTexParameteri (GLenum target, GLenum pname, GLfloat param)
 {
   if (pname != GL_GENERATE_MIPMAP)
     glTexParameteri (target, pname, param);
 }
 
 void
-cogl_wrap_glMaterialfv (GLenum face, GLenum pname, const GLfloat *params)
+_cogl_wrap_glMaterialfv (GLenum face, GLenum pname, const GLfloat *params)
 {
   /* FIXME: the GLES 2 backend doesn't yet support lighting so this
      function can't do anything */
 }
 
+void
+_cogl_wrap_glPointSize (GLfloat size)
+{
+  _COGL_GET_GLES2_WRAPPER (w, NO_RETVAL);
+
+  w->point_size = size;
+  w->dirty_uniforms |= COGL_GLES2_DIRTY_POINT_SIZE;
+}
+
 void
 _cogl_gles2_clear_cache_for_program (CoglHandle user_program)
 {
