@@ -37,6 +37,7 @@
 static void _cogl_shader_free (CoglShader *shader);
 
 COGL_HANDLE_DEFINE (Shader, shader);
+COGL_OBJECT_DEFINE_DEPRECATED_REF_COUNTING (shader);
 
 static void
 _cogl_shader_free (CoglShader *shader)
@@ -45,8 +46,6 @@ _cogl_shader_free (CoglShader *shader)
      released! Do that separately before this! */
   _COGL_GET_CONTEXT (ctx, NO_RETVAL);
   glDeleteShader (shader->gl_handle);
-
-  g_slice_free (CoglShader, shader);
 }
 
 CoglHandle
@@ -68,7 +67,6 @@ cogl_create_shader (CoglShaderType type)
 
   shader = g_slice_new (CoglShader);
   shader->gl_handle = glCreateShader (gl_type);
-  shader->type = type;
 
   return _cogl_shader_handle_new (shader);
 }
