@@ -38,6 +38,8 @@
 #include "../clutter-debug.h"
 #include "../clutter-main.h"
 
+#include "cogl/cogl-internal.h"
+
 #include <string.h>
 
 #include <glib.h>
@@ -105,6 +107,10 @@ clutter_event_source_new (ClutterBackend *backend)
   GSource *source = g_source_new (&event_funcs, sizeof (ClutterEventSource));
   ClutterEventSource *event_source = (ClutterEventSource *) source;
 
+#if GLIB_CHECK_VERSION (2, 25, 8)
+  g_source_set_name (source, "Clutter X11 Event");
+#endif
+
   event_source->backend = backend;
 
   return source;
@@ -454,6 +460,10 @@ event_translate (ClutterBackend *backend,
         }
     }
 
+  /* Pass the event through Cogl */
+  if (_cogl_xlib_handle_event (xevent) == COGL_XLIB_FILTER_REMOVE)
+    return FALSE;
+
   /* Do further processing only on events for the stage window (the x11
    * filters might be getting events for other windows, so do not mess
    * them about.
@@ -657,7 +667,7 @@ event_translate (ClutterBackend *backend,
       event->key.type = event->type = CLUTTER_KEY_PRESS;
       event->key.device =
         clutter_device_manager_get_core_device (manager,
-                                                CLUTTER_POINTER_DEVICE);
+                                                CLUTTER_KEYBOARD_DEVICE);
 
       translate_key_event (backend, event, xevent);
 
